#!/usr/bin/perl
# Credit card checking routines

package CCCheck;
use vars qw($VERSION $AUTOLOAD);
use strict;
use FreezeThaw;
use Cwd;
use File::KGlob qw(glob);

$VERSION = '1.48';

=head1 NAME

CCCheck.pm - credit card recognition module

=head1 SYNOPSIS

   use CCCheck;
   my $card=new CCCheck;
   # Card numbers can contain spaces or dashes
   $card->card_number("5473 0000 0000 0007");
   $card->expiry_date("05/99");
   # Optional.
   $card->start_date("11/91");
   # Alternatively, call $card->return_code to get a standard
   # DataCash return code - 1 for success, higher numbers for errors
   $card->check || die $card->error;
   print "Success for card ".$card->card_number."\n";
   print "This was a ".$card->issuer." card, bank was ".$card->bank."\n";

=head1 DESCRIPTION

CCCheck maintains a list of credit card details: how long credit card
number should be, whether they require an issue number or a start date,
their issuers and UK banks. It checks credit card numbers for validity
at the same time.

CCCheck provides the following methods.

=head2 new - create a CCCheck object

Supplied with no arguments, a single scalar parameter (a credit card number) or
a hash of arguments:

=over 4

=item * card_number

A credit card number (typically 13 to 19 digits long, no spaces or punctuation)

=item * expiry_date

An expiry date in the form mm/yy

=item * start_date

A start date (required by some cards) in the form mm/yy

=item * issue_number

An issue number (required by some cards), either one or two digits.

=back

Returns a CCCheck object. If the parameters are not filled in correctly,
CCCheck::error will return "Bad args" and the supplied arguments.

=cut

sub new {
	my $package=shift;
	my $self={};
	my %args;

	bless $self, $package;

	# Check what our arguments are, complain if we're supplied with nonsense
	if (scalar(@_)==1) {
		$args{card_number}=shift;
	} elsif (scalar(@_) % 2 ==0) {
		%args=@_;
	} else {
		$self->error("Bad args", @_);
	}
	
	$self->card_number($args{card_number});
	$self->expiry_date($args{expiry_date});
	$self->start_date($args{start_date});
	$self->issue_number($args{issue_number});
	return $self;
}

=head2 error - latest error reported by CCCheck

Returns the latest error reported by a CCCheck method, and an
(optional) list of arguments. When called in array context, returns
the error and arguments as a list. When called in string context,
returns in the form C<error>:C<argument1>,C<argument2> etc.

Currently only B<new> and B<check> return errors, so see
them for documentation.

=cut

sub error {
	my $self=shift;
	if (@_) {
		$self->{error}=shift;
		$self->{arguments}=[@_];
	} else {
		if ($self->arguments) {
			if (wantarray) { return ($self->{error}, @{$self->{arguments}}) }
			else { return $self->{error}.": ".join(", ",
																@{$self->{arguments}}) }
		} else {
			return $self->{error};
		}
	}
}

=head2 return_code - CCCheck error in number:args format

Returns 1 if there haven't been any errors, or a DataCash::External
error code if there have been. The mapping is as follows:

=over 4

=item 13 Bad arguments

=item 21 Bad type

=item 23 Malformed date I<or> Bad date

=item 24 No longer valid

=item 25 Bad checksum

=item 26 Wrong digits

=item 27 Bad issue

=item 28 Bad start date

=item 29 Not valid yet

=item 30 Start after expiry

=item 31 Two start dates

=item 33 Encrypted card

=back

=cut

sub return_code {
	my ($self)=@_;
	return 1 unless $self->error;
	my %error_codes=("Bad args" => 13, # init
						  "Bad type" => 21, # unrecognised card type - same thing
						  "Malformed date" => 23, # Not \d\d/\d\d
						  "No longer valid" => 24,
						  "Bad checksum" => 25,
						  "Wrong digits" => 26,	# Bad length
						  "Bad issue" => 27,
						  "Bad start date" => 28,
						  "Not valid yet" => 29,
						  "Bad date" => 23, # Out of range
						  "Start after expiry" => 30,	# Probably in future, not always
						  "Two start dates" => 31, # New
#						  "No issue please" => 32, # New as well
						  "Encrypted card" => 33,
						  );
	my $return_code=$error_codes{$self->{"error"}} || 10;
	$return_code.=":".join(",", @{$self->arguments})
		if ref($self->arguments) eq "ARRAY";
	return $return_code;
}

=head2 card_number - sets or reads the card number

Normally called by B<new>, unless you didn't supply a card number.
Once this method has been called, the following methods will return
useful results.

=cut

sub card_number {
	my ($self, $card)=@_;
	if ($card) {
		$self->{card_number}=$card;
		$self->{card_number} =~ tr/ -,.-//d;
		$self->read_card_details;
		$self->cache(0);
	} else {
		return $self->{card_number};
	}
}

=head2 length - the expected length of this type of credit card number

If the card number is of a recognised type, returns either a
number (currently either 13, 16, 18 or 19) or a range (13-19 for
Maestro cards). If it isn't recognised, length will return 0 instead.

=head2 want_start_date - whether we expect a start date

Returns 1 if we do, 0 if we don't.

=head2 want_issue_number - whether we expect an issue number

Return 1 if we expect a 1-digit issue number, 2 if we expect a
2-digit number (although a 1-digit number is accepted), 0
if we don't expect one (and we'll complain if we're given one).

=head2 issuer - which credit card company issued this card

Can be either any of the following:

=over 4

=item * VISA

=item * VISA Delta

=item * VISA UK Electron

=item * Mastercard

=item * Switch

=item * Solo

=item * JCB

=item * American Express

=item * Maestro

=item * Platima

=item * Aurore

=item * Diners

=back

B<issuers> will return a complete list.

=cut

# Accessor method for card_number, error, length,
# want_start_date and want_issue_number

sub AUTOLOAD {
    my $self=shift;
    my $name=$AUTOLOAD;
    return @_ ? $self->{$name}=shift : $self->{$name};
}

# Destroy all dbm handles when the whole thing dies
# This used to be a DESTROY method, but that meant that if you created
# two CCCheck objects and closed the first one, the second one would
# stop working

END {
	for (@CCCheck::dbmfiles) {
		dbmclose(%$_);
	}
}

=head2 bank - which bank issued this card

This will return things like "Bank of Scotland", "Midlands" and so on.
If you do not have the DBM files, it will be restricted to UK banks.

=cut

sub bank {
	my ($self)=@_;

	my @values=$self->bank_and_country;
	return shift @values;
}

=head2 country - the country this card was issued in

This will return things like "United Kingdom", "United States of
America", "Paraguay" and so on. If you do not have the DBM files,
it will return undef for foreign cards. It will also return undef
for American Express.

=cut

sub country {
	my ($self)=@_;

	my @values=$self->bank_and_country;
	return pop @values;
}

# Called by the bank and country methods, this returns an array of
# bank and country, looked up from either the built-in defaults or our
# external DBM files

sub bank_and_country {
	my ($self)=@_;

	# If we've been called recently, return the stored values
	if ($self->cache) {
		return @{$self->cache};
	}

	# First of all, if this is Amex, simply return American Express
	if ($self->issuer eq "American Express") {
		$self->cache(["American Express", undef]);
		return @{$self->cache};
	}

	# Get bank details - leave all the gory bits to somewhere deep down
	# in the source so we don't have to wade through it when we're looking
	# for the juicy bits. Also, we can potentially subclass it later.
	my $issuer_prefixes=$self->bank_details;
	# Get the prefix, searching deeper if this isn't a DBM file
	my $prefix=$self->match_range($self->card_number,
											$issuer_prefixes,
											$CCCheck::issuercache{$self->issuer."_dbm"} ?
											"bank" : "type");
	if (!$prefix) {
		$self->cache([undef,undef]);
		return undef,undef;
	}

	# This is either a straight string, or a frozen arrayref
	my $bank=$issuer_prefixes->{$prefix};
	my $country="United Kingdom";
	if ($bank =~ /^FrT/) {
		($bank, $country)=FreezeThaw::thaw($bank);
	}
	$self->cache([$bank, $country]);
	return $bank, $country;
}

# Check a card number for basic validity

=head2 valid_number - whether a credit card number is valid

Returns 1 if the credit card number passes a basic Luhn check, 0
if it fails, undef if there isn't a credit card number to check.

=cut

sub valid_number {
    my ($self)=@_;
    my ($total, $d);
	 my ($card_number)=$self->card_number;
	 return undef unless $card_number;
    $total = 0;
    while (length($card_number)) {
        $total += chop($card_number) +
            (($d = chop($card_number)) < 9 ? ($d * 2) % 9 : 9);
    }
    return !($total % 10);
}

=head2 check - check the credit card parameters

Checks a variety of things - whether the card number is valid, whether
we've specified the right kind of digits, whether the expiry date and,
optionally, start date are well formed and in the right time frame, etc.

Will return 1 if the card checks out, 0 if it fails a check; CCCheck::error
will then contain one of the following:

=over 4

=item * Bad checksum

Card number failed the basic Luhn check

=item * Bad type

This type of credit card is not recognised or supported

=item * Wrong digits

Card has wrong number of digits (argument: expected number of digits).

=item * Bad issue

Card has wrong number of digits in the issue number, or none was specified
(argument: expected number of digits).

=item * Malformed date

Expiry date was malformed (argument: date supplied).

=item * Bad start date

Start date was malformed or out of range.

=item * Bad date

Supplied month or year were out of range (expiry I<or> start date)

=item * Not valid yet

Card isn't valid yet (start date in the future)

=item * No longer valid

Card is no longer valid (expiry date in the past)

=item * Start after expiry

The specified start date is after the expiry date (foolish).

=item * Two start dates

A start date was supplied in both issue number and start date
fields, and it wasn't the same.

=item * No issue please

You supplied an issue number when none was requested.

=item * Card encrypted

The card number was encrypted and could not be checked.
All other checks performed OK, though.

=back

=cut

sub check {
	my ($self)=@_;

	my $encrypted=(length($self->card_number)>20 &&
						$self->card_number =~ /^[0-9a-f]+$/);

	# If the card number doesn't match a luhn check, complain and
	# quit now
	if (!$encrypted) {
		if (!$self->valid_number) {
			$self->error("Bad checksum");
			return 0;
		}

		# If this doesn't match any recognised type of credit card,
		# complain
		if (!$self->length) {
			$self->error("Bad type");
			return 0;
		}

		# If this has a different number of digits than we're expecting,
		# complain
		# This is a straight comparison if we just have a number, or
		# more complicated if we have a list or a range
		if (!$self->correct_length) {
			$self->error("Wrong digits", $self->length);
			return 0;
		}
	}

	# Check whether we've got a start date where we should have an issue
	# number, silently sort things out if we do
	if ($self->issue_number =~ m%(\d{1,2})/(\d{1,2})%) {
		if (!$self->start_date) {
			$self->start_date(sprintf("%02d/%02d", $1, $2));
			$self->issue_number("");
		} elsif ($self->start_date ne $self->issue_number) {
			$self->error("Two start dates", $self->start_date,
							 $self->issue_number);
			return 0;
		}
	}

	# If we have an issue number but didn't want one, clear it
	# silently
	if ($self->issue_number && !$self->want_issue_number) {
		$self->issue_number(undef);
#		$self->error("No issue please");
#		return 0;
	}

	# If there's the wrong number of digits or
	# the issue number is the wrong format, moan
	my $issue_digits=$self->want_issue_number;
	my $issue_number=$self->issue_number;
	if ($issue_digits) {
		if (!length($issue_number) ||
			 $issue_number =~ /\D/ ||
			 length(int($issue_number))>$issue_digits) {
			$self->error("Bad issue", $issue_digits);
			return 0;
		}
	}

	# Step through dates we require, sanity-check them
	my @dates=([$self->expiry_date, "Malformed date"]);
	my @fields_required=("expiry_date");
	push (@dates, [$self->start_date, "Bad start date"])
		if $self->want_start_date;
	my $date;
	date:
	foreach (@dates) {
		my ($date, $error)=@$_;
		# Barf if we can't make head or tail of it
		if ($date !~ m%^(\d{1,2})/(\d{2})$%) {
			# We can't handle this, so moan
			$self->error($error, $date);
			return 0;
		}
	}

	# If these dates are here, check them for validity
	# Work out our current, expiry and possibly start dates
	my $num_this_date=$self->numeric_date((gmtime)[4,5]);
	my $num_expiry_date=$self->numeric_date($self->expiry_date);
	my $num_start_date=($self->want_start_date ?
							  $self->numeric_date($self->start_date) : 0);
	# If either of these was bad, they'll be set to -1,
	# so check for that before parsing the expiry dates
	if ($num_start_date!=-1 && $num_expiry_date != -1) {
		if ($num_start_date<$num_expiry_date) {
			if ($num_this_date<$num_start_date) {
				$self->error("Not valid yet");
				return 0;
			}
			if ($num_this_date>$num_expiry_date) {
				$self->error("No longer valid");
				return 0;
			}
		} else {
			$self->error("Start after expiry");
			return 0;
		}
	} else {
		return 0;
	}

	# If this was an encrypted card, complain only now
	if ($encrypted) {
		$self->error("Encrypted card");
		return 0;
	}

	# If none of this has thrown up an error, the card must be OK then.
	return 1;
}

=head2 correct_length

Returns whether a card has a correct length.

=cut

sub correct_length {
	my ($self)=shift;
	if ($self->length !~ /\D/) {
		if (length($self->card_number) != $self->length) {
			return 0;
		}
	} else {
		my @lengths;
		for (split(",", $self->length)) {
			if (/-/) {
				s%-%..%;
				push (@lengths, eval($_));
			} else {
				push (@lengths, $_);
			}
		}
		my %lengths=map { $_, 1} @lengths;
		unless ($lengths{length($self->card_number)}) {
			return 0;
		}
	}
	return 1;
}

=head1 AUTHOR

Sam Kington (sam@datacash.com)

=cut

#### All routines after here are private to the module

# Returns a rough numeric date (number of months since 1900)
# Supplied with either "mm/yy" or "mm", "yy"
# Will return 0 if a value wasn't supplied

sub numeric_date {
	my $self=shift;
	if ($#_==0) {
		my ($origdate)=$_[0];
		my ($month, $year);
		($month, $year)=($origdate =~ m%^(\d{1,2})/(\d{2})$%);
		$year+=100 if $year<30;
		if ($month<=0 || $month >12 || $year >130) {
			$self->error("Bad date", $origdate);
			return -1;
		}
		return ($month-1)+$year*12;
	} else {
		return shift(@_)+shift(@_)*12;
	}
}

=head2 issuers - which types of card we know about

Returns an array of issuers we recognise, e.g. VISA, Switch etc.

=cut

sub issuers {
	return ("VISA", "VISA Delta", "VISA UK Electron", "VISA Purchasing",
			  "Switch", "Solo",
			  "JCB",
			  "Mastercard",
			  "American Express",
			  "Maestro",
			  "Platima",
			  "Aurore",
			  "Diners",
			  "transax",
			  );
}


# Private to the module; reads in data from a complex hash array
# and passes it to card_details, which sets 
#   length             => 0, 13, 16, 18 or 19 (0 if card is invalid range)
#   want_issue_number  => 0, 1, 2 (number of digits in issue number)
#   want_start_date    => 0, 1 (whether we require a start date)
#   issuer             => name of issuer

sub read_card_details {
	my ($self)=shift;
	my ($card_number) = $self->card_number;
	return unless $card_number;
	my ($range, $start, $end) = "";

	my %iin_prefixes=(
		# VISA 13 digit numbers (App. 5A), phase out by Jan 2001
#		"492946" => "VISA_16",
#		"492940" => "VISA_16",
#		"492900-45" => "VISA_13",
#		"492947-99" => "VISA_13",
#		"450335" => "VISA_13",
#		"450061" => "VISA_13",
		# 18-digit Switch numbers (App. 5H)
		"490302-09" => "Switch_18i1",
		"490335-40" => "Switch_18i1",
		"491174-82" => "Switch_18i1",
		"675938-40" => "Switch_18i1",

		# 18-digit Solo numbers (App. 5H)
		"633461" => "Solo_18i1",
		"633473" => "Solo_18i1",
		"633478" => "Solo_18i1",
		"633494" => "Solo_18i1",
		"633499" => "Solo_18i1",
		"676703" => "Solo_18i1",
		"676740" => "Solo_18i1",
		"676774" => "Solo_18i1",
		"676779" => "Solo_18i1",
		"676782" => "Solo_18i1",
		"676795" => "Solo_18i1",

		# 19-digit Switch numbers (App. 5I)
		"493600-99" => "Switch_19i1",
		"675901" => "Switch_19i1",
		"675918" => "Switch_19i1",
		"675950-62" => "Switch_19i1",
		"675998" => "Switch_19i1",

		# 19-digit Solo numbers (App. 5I)
		"633498" => "Solo_19i1",
		"676701" => "Solo_19i1",
		"676718" => "Solo_19i1",
		"676750-62" => "Solo_19i1",
		"676798" => "Solo_19i1",

		# 19-digit Switch numbers (App. 5J)
		"633301" => "Switch_19i1",
		"675905" => "Switch_19i1",

		# 19-digit Solo numbers (App. 5J)
		"633476" => "Solo_19i1",
		"633481" => "Solo_19i1",
		"676705" => "Solo_19i1",
		"676770" => "Solo_19i1",

		# 16-digit Switch numbers (App. 5D)
		"490525-29" => "Switch_16i2",

		# 16-digit Switch numbers (App. 5E)
		"564182" => "Switch_16i2",

		# 16-digit Solo numbers (App. 5E)
		"633495-97" => "Solo_16i2",
		"676706-07" => "Solo_16i2",

		# 16-digit Switch numbers (App. 5F)
		"491100-2" => "Switch_16i1",
		"633110" => "Switch_16i1",

		# 16-digit Solo numbers (App. 5F)
		"633490-93" => "Solo_16i1",

		# 16-digit Switch numbers (App. 5G)
		"633300" => "Switch_16s",
		"633302-49" => "Switch_16s",
		"675900" => "Switch_16s",
		"675902-04" => "Switch_16s",
		"675906-17" => "Switch_16s",
		"675919-37" => "Switch_16s",
		"675941-49" => "Switch_16s",
		"675963-97" => "Switch_16s",
		"675999" => "Switch_16s",

		# 16-digit Solo numbers (App. 5G)
		"633450-60" => "Solo_16s",
		"633462-72" => "Solo_16s",
		"633474-75" => "Solo_16s",
		"633477" => "Solo_16s",
		"633479-80" => "Solo_16s",
		"633482-89" => "Solo_16s",
		"676700" => "Solo_16s",
		"676702" => "Solo_16s",
		"676704" => "Solo_16s",
		"676708-17" => "Solo_16s",
		"676719-39" => "Solo_16s",
		"676741-49" => "Solo_16s",
		"676763-69" => "Solo_16s",
		"676771-73" => "Solo_16s",
		"676775-78" => "Solo_16s",
		"676780-81" => "Solo_16s",
		"676783-94" => "Solo_16s",
		"676796-97" => "Solo_16s",
		"676799" => "Solo_16s",

		# Phased out by October 2001 (App. 5G)
		# This could be either Switch or Solo, don't know (spec doesn't say)
		# I'll say Switch here but I could be wrong
		"633350-99" => "Switch_16s",

		# ATM-only cards (don't accept transactions)
		"490400-99" => "ATM_00",
		"492800-99" => "ATM_00",
# This is now a Japanese bank, apparently
#		"498600-99" => "ATM_00",
		"498700-99" => "ATM_00",
		# This is a Lloyds ATM card, apparently
		"492183" => "ATM_00",

		# VISA Delta 16-digit
		"446200-99" => "VISA Delta_16",
		"453978-79" => "VISA Delta_16",
		"454313" => "VISA Delta_16",
		"454432-35" => "VISA Delta_16",
		"454742" => "VISA Delta_16",
		"456725-45" => "VISA Delta_16",
		"465830-79" => "VISA Delta_16",
		"465901-50" => "VISA Delta_16",
		"490960-79" => "VISA Delta_16",
		"498824" => "VISA Delta_16",
		"492181-82" => "VISA Delta_16",

		# VISA UK Electron
		"491730-59" => "VISA UK Electron_16",
		"491880" => "VISA UK Electron_16",
		"484406-55" => "VISA UK Electron_16",
		"450875" => "VISA UK Electron_16",

		# Visa purchasing
		"448400-99" => "VISA Purchasing_16",
		"448500-99" => "VISA Purchasing_16",
		"448600-99" => "VISA Purchasing_16",
		"471500-99" => "VISA Purchasing_16",

		# Maestro (see also below)
		"300000" => "Maestro_13-19",
		"311079" => "Maestro_13-19",
		"367000" => "Maestro_13-19",

		# Carte Aurore (see also below)
		"501765-66" => "Aurore_19",
		"600522" => "Aurore_19",
		"712380-89" => "Aurore_19",

		# Transax (special case)
		"cheque" => "transax_6",

		# Platima
		"633591" => "Platima_16",

		# Standard VISA (starts with 4),
		# Mastercard (starts with 4, 51, 52, 53, 54 or 55),
		# Amex (starts with 34 or 37)
		# JCB (from 352800 to 358999)
		# Maestro (apart from already allocated ranges)
		# Aurore (starts with 100 or 200, plus already allocated ranges)
		"_VISA" => "VISA_13,16",
		"_Mastercard" => "Mastercard_16",
		"_Amex" => "American Express_15",
		"_JCB" => "JCB_16",
		"_Maestro" => "Maestro_13-19",
		"_Aurore" => "Aurore_19",
		"_Diners" => "Diners_14",
	);

	my $prefix=$self->match_range($card_number, \%iin_prefixes, "type");
	if ($prefix) {
		$self->card_details($prefix, \%iin_prefixes);
		return 1;
	} else {
		$self->error("Bad type");
		return 0;
	}
}

# Supplied with a card number and a hashref of card prefix => stuff
# (doesn't matter what the stuff is), return either the prefix that
# matches of 0 if nothing matched

sub match_range {
	my ($self, $card_number, $prefixes, $reqtype)=@_;
	my $iin=substr($card_number, 0, 6);

	# If we have an exact match, look up the stuff and return it
	if ($prefixes->{$iin} && $prefixes->{$iin} ne "") {
		return $iin;
	} else {
		# We'll look for ranges. First of all, try the wide and narrow
		# ranges
		my @prefixes;
		my $iin4=substr($card_number, 0, 4);
		$self->add_prefix($iin4."00-99", $prefixes, \@prefixes);
		my $digit5=substr($card_number, 4, 1);
		$self->add_prefix($iin4."${digit5}0-${digit5}9", $prefixes, \@prefixes);
		# If we haven't found anything, grep the keys (this could take
		# a *hell* of a long time in a DBM file)
		if (!@prefixes && $reqtype ne "bank") {
			@prefixes = grep(/^$iin4/, keys %$prefixes);
		}
		my ($range, $start, $end);
		foreach $range (@prefixes) {
			# Ranges will be either 123456, 123456-7 or 123456-78
			if ($range =~ /(\d+)-(\d+)/) {
				$start=$1;
				# This comes out as, roughly, $end =~ s%(\d{2})$%78%
				eval ('($end=$start)=~s%(\d{'.length($2).'})$%'.$2.'%;');
			} else {
				$start=$end=$range;
			}
			# If we have a match, pull in the rest of the stuff and quit
			if ($iin>=$start && $iin<=$end) {
				return $range;
			}
		}
	}

	# If we didn't find a match, check wider ranges
	if ($card_number =~ /^4/) {
		return "_VISA";
	} elsif ($card_number =~ /^(51|52|53|54|55)/) {
		return "_Mastercard";
	} elsif ($card_number =~ /^(34|37)/) {
		return "_Amex";
	} elsif ($card_number =~ /^(50|56|57|58|59|6)/) {
		return "_Maestro";
	} elsif ($card_number =~ /^35(28|29|[345678])/) {
		return "_JCB";
	} elsif ($card_number =~ /^(100|200)/) {
		return "_Aurore";
	} elsif ($card_number =~ /^(30[0-5]|36|38)/) {
		return "_Diners";
	}

	# We're still here, so it didn't match: return null values
	return 0;
}

sub add_prefix {
	my ($self, $range, $hash, $list)=@_;
	push(@$list, $range) if $hash->{$range};
}

# Called by credit_card_details: actually generates the anonymous hash
# to be returned

sub card_details {
	my ($self, $key, $iin_prefixes)=@_;
	my ($wantissue, $wantstart);

	# Look up the codes for this range
	my $val=$iin_prefixes->{$key};
	# Extract issuer and length, remove them from the string
	my ($issuer, $length)=($val =~ /^(.*?)_([\d,-]+)/);
	$val =~ s%${issuer}_${length}%%;

	if ($val =~ /i(\d)$/) {
		$wantissue=$1;
		$wantstart=0;
	} elsif ($val =~ /s$/) {
		$wantissue=0;
		$wantstart=1;
	}

	$self->length($length);
	$self->want_issue_number($wantissue);
	$self->want_start_date($wantstart);
	$self->issuer($issuer);

}

# Called by bank, sets up a hashref of prefixes and corresponding banks
# Checks issuer first to avoid returning too much detail
=pod
=item * VISA
=item * VISA Delta
=item * VISA UK Electron
=item * Mastercard
=item * Switch
=item * Solo
=item * JCB
=item * Maestro
=cut

sub bank_details {
	my ($self)=@_;

	my $issuer=$self->issuer;
	return {} if !$issuer;
	my $details;
	# If we've already stored this type of card's information,
	# return the cached version immediately
	if ($details=$CCCheck::issuercache{$issuer}) {
		return $details;
	}
	else { $details={} }
	# Keep track of whether this is a DBM file or not
	$CCCheck::issuercache{$issuer."_dbm"}=0;
	if ($issuer eq "VISA") {
		$details=$self->open_dbm($issuer) || {
			"458056"=>"Abbey National B/Soc",
			"491795"=>"Abbey National B/Soc",
			"453975"=>"Adam & Company",
			"456000"=>"Adam & Company",
			"450924"=>"Aib (GB Only)",
			"450925"=>"Aib (GB Only)",
			"454364"=>"Aib (GB Only)",
			"493847"=>"Aib (NI Only)",
			"493848"=>"Aib (NI Only)",
			"480925"=>"Aib (NI Only)",
			"456089"=>"Allied Trust Bank",
			"450831"=>"Associates Capital",
			"451133"=>"Associates Capital",
			"494017"=>"Banco De Bilbao",
			"494018"=>"Banco De Bilbao",
			"490159"=>"Banco De Sabadell",
			"492042"=>"Banco De Sabadell",
			"458051-52"=>"Banco Exterior De Espana",
			"494315"=>"Bank Of Austria",
			"440791-92"=>"Bank Of Austria",
			"451171"=>"Bank Of China",
			"451172"=>"Bank Of China",
			"455273"=>"Bank Of Cyprus",
			"414936"=>"Bank Of Cyprus",
			"492083"=>"Bank Of Ireland",
			"498834"=>"Bank Of Ireland",
			"479741-42"=>"Bank Of Ireland",
			"479743"=>"Bank Of Ireland",
			"450783"=>"Bank Of Scotland",
			"456072"=>"Bank Of Scotland",
			"491605"=>"Bank Of Scotland",
			"491700"=>"Bank Of Scotland",
			"456073"=>"Bank Of Scotland",
			"492901-2"=>"Barclays Bank Plc",
			"492906"=>"Barclays Bank Plc",
			"492928-30"=>"Barclays Bank Plc",
			"492939"=>"Barclays Bank Plc",
			"492937"=>"Barclays Bank Plc",
			"492938"=>"Barclays Bank Plc",
			"492946"=>"Barclays Bank Plc",
			"492950-62"=>"Barclays Bank Plc",
			"492966"=>"Barclays Bank Plc",
			"492968-69"=>"Barclays Bank Plc",
			"492970-98"=>"Barclays Bank Plc",
			"471532"=>"Barclays Bank Plc",
			"430532"=>"Barclays International",
			"454424"=>"Beneficial Bank",
			"450891"=>"Beneficial Bank",
			"492207"=>"Beneficial Bank",
			"454774"=>"Beneficial Bank",
			"491233"=>"Beneficial Bank",
			"477595"=>"Beneficial Bank",
			"479700"=>"Beneficial Bank",
			"477596"=>"Capital One",
			"477597"=>"Capital One",
			"479293-95"=>"Cater Allen",
			"453976"=>"C Hoare & Co",
			"440669"=>"Citibank",
			"491001"=>"Citibank",
			"491002"=>"Citibank",
			"445984"=>"Citibank",
			"403542"=>"Citibank",
			"403573"=>"Citibank",
			"454365-66"=>"Clydesdale Bank",
			"454495"=>"Co-Operative Bank",
			"455025"=>"Co-Operative Bank",
			"455206"=>"Co-Operative Bank",
			"498823"=>"Co-Operative Bank",
			"455572"=>"Co-Operative Bank",
			"455783"=>"Coutts & Co",
			"426353"=>"Coutts & Co",
			"403543"=>"Coutts & Co",
			"494006"=>"Credit Lyonnais",
			"491232"=>"Cyprus Popular Bank",
			"415467"=>"Cyprus Popular Bank",
			"455232"=>"Duncan Lawrie",
			"471577"=>"First Personal Bank",
			"479244-45"=>"First Personal Bank",
			"451113"=>"Frizzell Banking Services",
			"454769"=>"Girobank",
			"455653"=>"Girobank",
			"455299"=>"Girobank",
			"493844"=>"Girobank",
			"455647"=>"Halifax",
			"451153"=>"Hfc Bank",
			"477928"=>"Hfc Bank",
			"440710"=>"Hfc Bank",
			"447583"=>"Hfc Bank",
			"449355"=>"Hfc Bank",
			"464496"=>"Hfc Bank",
			"494348"=>"Hfc - Vauxhall Motors",
			"494067"=>"Jardine Fleming",
			"494079-80"=>"Jardine Fleming",
			"430563"=>"Kleinwort Benson",
			"454103"=>"Leeds Permanent",
			"492183"=>"Lloyds",
			"448445"=>"Lloyds",
			"491600"=>"MBNA",
			"491601"=>"MBNA",
			"454312"=>"Midland Bank",
			"454361"=>"Midland Bank",
			"454638"=>"Midland Bank",
			"455592"=>"Midland Bank",
			"455598"=>"Midland Bank",
			"455796"=>"Midland Bank",
			"471507"=>"Midland Bank",
			"494120"=>"Midland Bank",
			"493833"=>"National & Provincial",
			"459970"=>"National & Provincial",
			"494070"=>"National Bank Of Kuwait",
			"402193"=>"Nat West Bank",
			"454438"=>"Nat West Bank",
			"454420"=>"Nat West Bank",
			"453929"=>"Nat West Bank",
			"453937"=>"Nat West Bank",
			"471505"=>"Nat West Bank",
			"426393"=>"Nat West Bank",
			"477591"=>"Nat West Bank",
			"445958"=>"Nat West Bank",
			"449352"=>"Nationwide B/Soc",
			"449353"=>"Nationwide B/Soc",
			"454369"=>"Northern Bank",
			"490918"=>"Northwest Securities",
			"447963"=>"Peoples Bank",
			"447964"=>"Peoples Bank",
			"447965"=>"Peoples Bank",
			"450859"=>"Robert Fleming",
			"450901"=>"Robert Fleming",
			"450999"=>"Robert Fleming",
			"454639"=>"Robert Fleming",
			"455207"=>"Robert Fleming",
			"456875"=>"Robert Fleming",
			"400880"=>"Rbs Advanta",
			"445974"=>"Rbs Advanta",
			"454305"=>"Royal Bank Of Scotland",
			"491007"=>"Royal Bank Of Scotland",
			"456849"=>"Royal Bank Of Scotland",
			"448405"=>"Royal Bank Of Scotland",
			"430567"=>"Royal Bank Of Scotland",
			"491834"=>"Royal Bank Of Scotland",
			"477916"=>"Sainsburys Bank",
			"477917"=>"Sainsburys Bank",
			"450860"=>"Standard Chartered",
			"450823"=>"TSB Trustcard",
			"450874"=>"TSB Trustcard",
			"450897"=>"TSB Trustcard",
			"454634"=>"Ulster Bank",
			"471558"=>"Ulster Bank",
			"401840"=>"Woolwich",
			"401841"=>"Woolwich",
			"450906"=>"Yorkshire Bank",
			"454443"=>"Yorkshire Bank",
		};
	} elsif ($issuer eq "VISA Delta") {
		$details={
			"454742"=>"Abbey National B.Soc",
			"454434"=>"Aib (NI)",
			"456735"=>"Alliance & Leicester",
			"490975"=>"Arbuthnot  Latham",
			"456739"=>"Bank Of Cyprus",
			"453978"=>"Barclays-Connect",
			"453979"=>"Barclays-Connect",
			"456733"=>"Benefical Bank",
			"456727"=>"C Hoare & Co",
			"456738"=>"Citibank",
			"456744"=>"Citibank",
			"498824"=>"Co Operative Bank",
			"456728"=>"Credit Lyonnais",
			"456726"=>"Girobank",
			"456731-32"=>"Guinness Mahon",
			"456737"=>"Guinness Mahon",
			"456730"=>"Harrods Bank",
			"490974"=>"Kleinwort Benson",
			"492181-82"=>"Lloyds Bank",
			"490964"=>"Lloyds Bank",
			"490976"=>"National Bank Of Kuwait",
			"456743"=>"National & Provincial",
			"454313"=>"Nationwide",
			"490969"=>"Northern Rock B Soc",
			"490978-79"=>"Standard Bank",
			"490960-63"=>"Standard Chartered Bank",
			"490965-68"=>"Standard Chartered Bank",
			"490970"=>"Standard Chartered Bank",
			"490971-73"=>"Standard Chartered Bank",
			"454432"=>"TSB",
			"454433"=>"TSB",
			"454435"=>"TSB",
			"456740-42"=>"United Bank Of Kuwait",
			"456725"=>"Woolwich B Soc",
		};
	} elsif ($issuer eq "VISA UK Electron") {
		$details={
			"491731"=>"Abbey National",
			"491880"=>"Bank of Cyprus",
			"491748-49"=>"Barclays",
			"450875"=>"Co-op",
			"491737"=>"Northern Rock ",
			"491733"=>"Woolwich",
		};
	} elsif ($issuer eq "Mastercard") {
		$details=$self->open_dbm($issuer) || {
			"548507"=>"Abbey National",
			"545664"=>"Alliance & Leicester",
			"545994"=>"Alliance & Leicester",
			"545995"=>"Alliance & Leicester",
			"542786"=>"Allied Irish Banks",
			"543046"=>"Allied Irish Banks",
			"543269"=>"Bank Of Ireland",
			"543268"=>"Bank Of Ireland",
			"546259"=>"Bank Of Ireland",
			"546260"=>"Bank Of Ireland",
			"546261"=>"Bank Of Ireland",
			"546262"=>"Bank Of Ireland",
			"542195"=>"Bank Of Scotland",
			"540152"=>"Bank Of Scotland",
			"540153"=>"Bank Of Scotland",
			"530125"=>"Barclays Bank",
			"546096"=>"Capital One",
			"546097"=>"Capital One",
			"542061"=>"Citibank",
			"542064"=>"Citibank",
			"540453"=>"Clydesdale Bank",
			"543477"=>"Clydesdale Bank",
			"547203"=>"Clydesdale Bank",
			"547353"=>"Clydesdale Bank",
			"540964"=>"Coutts & Co",
			"543474"=>"Coutts & Co",
			"545100"=>"Coutts & Co",
			"543276"=>"Frizzell Banking Services",
			"546895"=>"GE Capital",
			"547101"=>"GE Capital",
			"547108"=>"GE Capital",
			"547114"=>"GE Capital",
			"547116"=>"GE Capital",
			"547152"=>"GE Capital",
			"547735"=>"GE Capital",
			"547736"=>"GE Capital",
			"547817-43"=>"GE Capital",
			"547847"=>"GE Capital",
			"547848"=>"GE Capital",
			"542499"=>"Girobank",
			"541914"=>"HFC",
			"542854"=>"HFC",
			"543131"=>"HFC",
			"544576"=>"HFC",
			"545852"=>"HFC",
			"548041"=>"HFC",
			"540438-39"=>"Lloyds Bank",
			"540440"=>"Lloyds Bank",
			"540463"=>"Lloyds Bank",
			"544182"=>"MBNA",
			"545043"=>"MBNA",
			"543458"=>"Midland Bank",
			"547347"=>"Midland Bank",
			"543460"=>"Midland Bank",
			"543461"=>"Midland Bank",
			"543468"=>"Natwest",
			"543557"=>"Natwest",
			"543699"=>"Natwest",
			"547367"=>"Natwest",
			"543556"=>"Natwest",
			"545460"=>"Natwest",
			"540454"=>"Northern Bank",
			"543478"=>"Northern Bank",
			"547354"=>"Northern Bank",
			"546297"=>"People's Bank",
			"546298"=>"People's Bank",
			"547872"=>"People's Bank",
			"542851"=>"Royal Bank Of Scotland",
			"543480"=>"Royal Bank Of Scotland",
			"547356"=>"Royal Bank Of Scotland",
			"543482"=>"Royal Bank Of Scotland",
			"546016"=>"Royal Bank Of Scotland",
			"543484"=>"Royal Bank Of Scotland",
			"545710"=>"Royal Bank Of Scotland",
			"546019"=>"Royal Bank Of Scotland",
			"542435"=>"Save & Prosper",
			"542502"=>"TSB ",
			"543470"=>"Ulster Bank",
			"541658"=>"Woolwich",
		};
	} elsif ($issuer eq "Switch") {
		$details={
			"490302-09"=>"Midland Bank",
			"490335-40"=>"Midland Bank",
			"490525-29"=>"Royal Bank of Scotland",
			"491100-02"=>"Bank of Scotland",
			"491174-82"=>"Clydesdale Bank (including Northern Bank)",
			"493600-99"=>"NatWest (including Ulster Bank)",
			"564182"=>"Halifax",
			"633110"=>"Bank of Scotland",
			"633300"=>"Midland Bank",
			"633301"=>"Yorkshire Bank",
			"633302-04"=>"Bank of Scotland",
			"633311"=>"Royal Bank of Scotland",
			"633312"=>"Bank of Scotland",
			"675901"=>"NatWest (including Ulster Bank)",
			"675905"=>"Yorkshire Bank",
			"675918"=>"NatWest (including Ulster Bank)",
			"675938-40"=>"Midland Bank",
			"675950-62"=>"NatWest (including Ulster Bank)",
			"675963-70"=>"Royal Bank of Scotland",
			"675971-73"=>"Bank of Scotland",
			"675982"=>"Clydesdale Bank (including Northern Bank)",
			"675995"=>"Clydesdale Bank (including Northern Bank)",
			"675998"=>"NatWest (including Ulster Bank)",
		};
	} elsif ($issuer eq "Solo") {
		$details={
			"633450-53"=>"Royal Bank of Scotland",
			"633454-57"=>"Bank of Scotland",
			"633461"=>"Midland Bank",
			"633473"=>"Northern Bank",
			"633476"=>"Yorkshire Bank",
			"633478"=>"Clydesdale Bank",
			"633481"=>"Yorkshire Bank",
			"633490-93"=>"Bank of Scotland",
			"633494"=>"Clydesdale Bank",
			"633495-97"=>"Halifax",
			"633498"=>"Ulster Bank",
			"633499"=>"Clydesdale Bank",
			"676701"=>"NatWest",
			"676703"=>"Midland Bank",
			"676705"=>"Yorkshire Bank",
			"676706-07"=>"Halifax",
			"676708-11"=>"Royal Bank of Scotland",
			"676712-15"=>"Bank of Scotland",
			"676718"=>"NatWest",
			"676740"=>"Midland Bank",
			"676750-62"=>"NatWest",
			"676770"=>"Yorkshire Bank",
			"676774"=>"Clydesdale Bank",
			"676779"=>"Clydesdale Bank",
			"676782"=>"Clydesdale Bank",
			"676795"=>"Clydesdale Bank",
			"676798"=>"Ulster Bank",
		};
	}
	# Cache this for later use
	$CCCheck::issuercache{$issuer}=$details;
	return $details;
}

sub open_dbm {
	my ($self, $name)=@_;
	my $cwd=cwd();
	chdir $self->filebase;
	my %array;
	dbmopen(%array, lc($name), 0644) || return 0;
	push (@CCCheck::dbmfiles, \%array);
	chdir $cwd;
	$CCCheck::issuercache{$name."_dbm"}=1;
	return \%array;
}

# Return the directory our DB files are in

sub filebase {
	my ($self)=@_;
	if (!defined $CCCheck::filebase) {
		my $cwd=cwd();
		directory:
		for (@INC) {
			$_ eq "." ? chdir $cwd : chdir $_;
			next directory if !-e "CCCheck.pm";
			if (-e "CCCheck" && -d "CCCheck") {
				chdir "CCCheck";
				if (glob("mastercard*") && glob("visa*")) {
					$CCCheck::filebase=cwd;
					last directory;
				}
			}
 			if (glob("mastercard*") && glob("visa*")) {
				$CCCheck::filebase=cwd;
				last directory;
			}
		}
		chdir($cwd);
	}
	return $CCCheck::filebase;
}

1;
