#!/usr/bin/perl

use strict;
use DataCash::External;
use Getopt::Long;
use CCCheck;

# Define our defaults
my %options=(vtid => 21859999,
				 client_pass => "fred",
				 blowfish => "0123456789ABCDEF",
				 num => "5473000000000007",
				 exp => "04/03",
				 issue => undef,
				 start => undef,
				 amount => "1.00",
				 currency => "GBP",
				 type => "auth",
				 host => "new-auth.datacash.com",
				 port => 8001,
				 );

# Read in our command line options
my %params;
GetOptions(\%params,
			  qw(help|h
				  vtid=i client_pass=s blowfish=s
				  num=s exp=s issue=s start=s
				  amount=s currency=s ref=i
				  type=s host=s port=i)) || die usage();
# If this is a help request, die now.
die usage() if $params{help};

# Replace any unspecified options with the defaults
for (keys %options) {
	$params{$_} ||= $options{$_};
}

# Clean up parameters slightly
for (keys %params) {
	$params{$_} =~ s%^\s+%%;
	$params{$_} =~ s%\s+$%%;
}
for (qw(num issue)) {
	$params{$_} =~ s%\D%%g;
}
for (qw(exp start)) {
	$params{$_} =~ s%[^0-9/]%%g;
}
$params{ref} ||= time | $$;

# Check whether we need an issue number or a start date, remove
# extraneous parameters, and insert the remaining value (if any) into
# issue
my $card=new CCCheck($params{num});
delete $params{issue} unless $card->want_issue_number;
delete $params{start} unless $card->want_start_date;
$params{issue}=$params{issue} || $params{start};
delete $params{start};

# Set up a DataCash::External, and remove these parameters from our list
my $ext=new DataCash::External($params{vtid});
$DataCash::Blowfish=$params{blowfish};
$ext->{defaults}{back_addr}=$params{host};
$ext->{defaults}{back_sock}=$params{port};
for (qw(vtid blowfish host port)) { delete $params{$_} }

# Whatever remains is to be passed to the DataCash::External via the
# normal manner
for (keys %params) {
	$@=undef;
	eval ("\$ext->$_('".$params{$_}."')");
	warn "$@\n" if $@;
}

my $result;
eval {
	$result=$ext->send;
};
$result ||= "3:server did not respond";
chomp $result;
print $result,"\n";


sub usage {
	return <<USAGE;
Usage: datacash.pl --help
or     datacash.pl --num <card number>
                   --exp <mm/yy>
                   [--start <mm/yy> | --issue <issue number>]
                   --amount <value of transaction>
                   --currency <ISO currency code>
                   --vtid <vTID number>
                   --password <password>
                   --blowfish <encryption key>
                   --ref <6-12 digit reference number>
                 	[--type (auth|pre|refund|fulfill)]
                  [--host <host name>] [--port <port number>]

Supplied with credit card details, sends a request to the DataCash
payment server and returns the raw DataCash response, which is a
series of arguments separated by colons (':').

Entry arguments are:
   --num          Card number (can include punctuation or spaces)
   --exp          Expiry date (mm/yy format)
   --start        Start date (mm/yy format), Switch cards only
   --issue        Issue number, Switch cards only
                  Only specify one of these
   --amount       Value of transaction, fully specified, e.g. 12.00
   --currency     Currency, e.g. GBP, USD etc.
   --ref          Unique reference number (6-12 digits)
   --vtid         vTID number (defaults to 21859999)
   --client_pass  Password (defaults to fred)
   --blowfish     Blowfish encryption key (defaults to 0123456789ABCDEF)
   --type         Type of transaction - auth, pre, refund or fulfill
                  Defaults to "auth"
   --host         Host name of the server (defaults to new-auth.datacash.com)
   --port         Port number of the server (defaults to 8001)

A random reference number will be generated if none is supplied.

For a successful transaction, or a bank decline, the response will
look like e.g.
1:024543:907927010:2185999900000691:VISA:Barclays Bank Plc:United	Kingdom

The arguments are:
* The DataCash return code (see
  http://www.datacash.com/framed.cgi/stronghold/returncodes.html for a
  complete list). In this case, '1' (success).
* The authorisation code, or bank response or other informative
  message.
* The time of the transaction (Unix timestamp - number of seconds
  since 1st January 1970 GMT)
* DataCash internal reference number
* The type of card
* The issuing bank
* The country of issue

Other return codes will include fewer arguments, typically just return
code and optional extra argument (e.g. 36: speed limit).
USAGE
}
